// This is a part of the GWC Library
// Copyright (C) 2001-2002 Aircom software
// All rights reserved.
//
// This source code can be used, distributed or modified
// free of charge only if this header copyright is not removed.
// In other cases, contact us at:
// web:   http://www.aircom.org
// email: info@aircom.org
//

#include "stdafx.h"
#include "GWCColorComponentEditCtrl.h"
#include "GWCGradientColorWnd.h"
#include "GWCColorComponentSet.h"
#include "memdc.h"
#include "GWCColorFunc.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


// Message sent to the parent window to notify it that the value of this component
// has changed.
UINT WM_GWCCOLORCOMPEDIT_CHANGED = ::RegisterWindowMessage(_T("WM_GWCCOLORCOMPEDIT_CHANGED"));



//=====================================================
// Description:
// Constructs a new GWCColorComponentEditCtrl instance.
//=====================================================
GWCColorComponentEditCtrl::GWCColorComponentEditCtrl() : m_buttonPushed(false), m_value(0),
	m_externalEditUpdate(false), m_dragInProgress(false), m_ownerSet(NULL), m_gradientChooser(NULL),
	m_buttonFocus(false)
{
	// The window class is registered right now to be used in dialog boxes or
	// other windows.

	RegisterWindowClass();
}

//==================================================================
// Description:
// At destruction time, cleans up some memory used by this instance.
//==================================================================
GWCColorComponentEditCtrl::~GWCColorComponentEditCtrl()
{
	// May be NULL if the right button has not been clicked but delete won't complain.
	delete m_gradientChooser;
}


BEGIN_MESSAGE_MAP(GWCColorComponentEditCtrl, CWnd)
	//{{AFX_MSG_MAP(GWCColorComponentEditCtrl)
	ON_WM_PAINT()
	ON_WM_ERASEBKGND()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_VSCROLL()
	ON_WM_KILLFOCUS()
	ON_WM_SETFOCUS()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//=================================================================================
// Description:
// Dynamically creates the window handle of this control. Don't call it when this
// control is used in a dialog box.
//
// Parameters:
// dwStyle - Specifies the window style attributes.
// rect - The size and position of the window, in client coordinates of pParentWnd.
// pParentWnd - The parent window.
// nID - The ID of this child window.
//
// Returns:
// Nonzero if successful, otherwise 0.
//=================================================================================
BOOL GWCColorComponentEditCtrl::Create(DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID)
{
	// We use the window class already registered
	return CWnd::CreateEx(0,COLORCOMPEDITCTRL_CLASSNAME, NULL, dwStyle, rect, pParentWnd, nID);
}

//===========================================================================
// Description:
// Registers a window class, so that a client can add a custom control in the
// dialog editor.
//
// Returns:
// TRUE if successfull, FALSE otherwise.
//===========================================================================
BOOL GWCColorComponentEditCtrl::RegisterWindowClass()
{
    WNDCLASS wndcls;
    HINSTANCE hInst = AfxGetInstanceHandle();

	// Already registered ?
    if (!(::GetClassInfo(hInst, COLORCOMPEDITCTRL_CLASSNAME, &wndcls)))
    {
        // otherwise we need to register a new class
        wndcls.style            = CS_CLASSDC | CS_SAVEBITS;
        wndcls.lpfnWndProc      = ::DefWindowProc;
        wndcls.cbClsExtra       = wndcls.cbWndExtra = 0;
        wndcls.hInstance        = hInst;
        wndcls.hIcon            = NULL;
        wndcls.hCursor          = AfxGetApp()->LoadStandardCursor(IDC_ARROW);
        wndcls.hbrBackground    = NULL;
        wndcls.lpszMenuName     = NULL;
        wndcls.lpszClassName    = COLORCOMPEDITCTRL_CLASSNAME;

        if (!AfxRegisterClass(&wndcls))
        {
            AfxThrowResourceException();
            return FALSE;
        }
    }

    return TRUE;
}

//========================================================================
// Description:
// Called by the framework after creation when the window handle is valid.
//========================================================================
void GWCColorComponentEditCtrl::PreSubclassWindow() 
{
	CWnd::PreSubclassWindow();

	// Compared to a standard CEdit height, we increase it by one for the best drawing effect.
	// In the dialog editor, just create a custom control the same height as a CEdit control.
	// If used dynamically in a window, create it with a height equal to 23.
	CRect wRect;
	GetWindowRect(&wRect);
	wRect.bottom++;
	GetParent()->ScreenToClient(&wRect);
	MoveWindow(&wRect,FALSE);

	// There are some child controls so use WS_CLIPCHILDREN
	ModifyStyle(0,WS_CLIPCHILDREN);

	// Add the same border effect as a CEdit
	ModifyStyleEx(0,WS_EX_CLIENTEDGE,SWP_FRAMECHANGED);

	// Creation of children
	//---------------------

	CRect rect;
	GetClientRect(&rect);

	// Get the standard GUI font
	CFont font;
	HFONT hFont = (HFONT)::GetStockObject(DEFAULT_GUI_FONT);
	font.Attach(hFont);

	// Create the CEdit
	CRect editRect = rect;
	editRect.DeflateRect(1,1,24,4);
	m_edit.Create(WS_CHILD|WS_VISIBLE,editRect,this,1);
	m_edit.SetWindowText(_T("0"));
	m_edit.SetFont(&font);

	// Create the Spin
	CRect spinRect = rect;
	spinRect.bottom -= 4;
	spinRect.right -= 12;
	spinRect.left = spinRect.right - 12;
	m_spin.Create(WS_CHILD|WS_VISIBLE|UDS_ARROWKEYS|UDS_NOTHOUSANDS,spinRect,this,2);
	m_spin.MoveWindow(spinRect,FALSE);
	m_spin.SetRange(0,255);

	UDACCEL uda = { 1, 1 };
	m_spin.SetAccel(1,&uda);
}

//===============================================
// Description:
// Called by the framework to paint this control.
//===============================================
void GWCColorComponentEditCtrl::OnPaint() 
{
	// Create a memory DC for a non flickering draw
	CPaintDC paintDC(this);
	CMemDC dc(&paintDC);

	CRect rect;
	GetClientRect(&rect);

	// Draw background
	dc.FillSolidRect(&rect,::GetSysColor(COLOR_BTNFACE));

	// Draw the bottom color bar (colored part)
	CRect colorRect = GetColorRect();
	CRect rect1 = colorRect;
	rect1.right = rect1.left + Round((double)(rect1.Width() * m_value) / 255.0);
	dc.FillSolidRect(&rect1,m_colorBar);

	// Draw the bottom color bar (black part)
	CRect rect2 = colorRect;
	rect2.left = rect1.right;
	if (rect2.Width())
		dc.FillSolidRect(&rect2,RGB(0,0,0));

	// Draw false button
	CRect buttonRect = GetButtonRect();
	dc.DrawFrameControl(&buttonRect,DFC_BUTTON,DFCS_BUTTONPUSH|(m_buttonPushed ? DFCS_PUSHED : 0));
	dc.SelectStockObject(BLACK_PEN);
	int minx = 3 + (m_buttonPushed ? 1 : 0);
	int miny = 5 + (m_buttonPushed ? 1 : 0);
	dc.MoveTo(buttonRect.left+minx,buttonRect.top+miny);
	dc.LineTo(buttonRect.left+minx+5,buttonRect.top+miny);
	dc.MoveTo(buttonRect.left+minx+1,buttonRect.top+miny+1);
	dc.LineTo(buttonRect.left+minx+4,buttonRect.top+miny+1);
	dc.MoveTo(buttonRect.left+minx+2,buttonRect.top+miny+2);
	dc.LineTo(buttonRect.left+minx+3,buttonRect.top+miny+2);
	dc.MoveTo(buttonRect.left+minx,buttonRect.top+miny+4);
	dc.LineTo(buttonRect.left+minx+5,buttonRect.top+miny+4);
}

//===============================================================
// Description:
// Called before OnPaint to erase the background of this control. 
//
// Parameters:
// pDC - A pointer to the device context used for drawing.
//
// Returns:
// TRUE so that the parent won't paint the background.
//===============================================================
BOOL GWCColorComponentEditCtrl::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;
}

//===============================================================================
// Description:
// Called by the framework when the left mouse button is pressed in this control.
//
// Parameters:
// nFlags - Not used.
// point - Client coordinates of the mouse cursor.
//===============================================================================
void GWCColorComponentEditCtrl::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// Two actions can be done on this control when left cliking :
	// 1. A drag begins in the color bar
	// 2. The right button is clicked

	// Get the rectangles of the button and the color bar
	CRect buttonRect = GetButtonRect();
	CRect colorRect = GetColorRect();

	// If we are in the button
	if (buttonRect.PtInRect(point))
	{
		// The button must be redrawn with a pushed effect
		m_buttonPushed = true;
		RedrawWindow(&buttonRect,NULL);

		// Create the popup window to see the gradient color bar
		if (m_gradientChooser == NULL)
		{
			m_gradientChooser = new GWCGradientColorWnd(this);
			m_gradientChooser->Create(WS_POPUP,this,0);
		}

		// Show the gradient color bar
		point.y = colorRect.bottom + 1;
		ClientToScreen(&point);

		m_gradientChooser->Show(point);

		// Set the focus to this control
		m_buttonFocus = true;
		SetFocus();

		// Capture the mouse to drag a cursor in the popup color bar
		SetCapture();
	}

	// If we are in the bottom color bar
	else if (colorRect.PtInRect(point))
	{
		m_dragInProgress = true;

		// Calculate a value based on the location of the click
		m_value = Round((double)(point.x-colorRect.left)/colorRect.Width()*255.);

		// Update the edit control with this value
		CString str;
		str.Format(_T("%d"),m_value);
		m_edit.SetWindowText(str);

		// Redraw the control to update the color bar
		RedrawWindow(colorRect,NULL);

		// Capture the mouse to drag it in the color bar
		SetCapture();
	}

	CWnd::OnLButtonDown(nFlags, point);
}

//===================================================================
// Description:
// Called by the framework when the mouse is moved over this control.
//
// Parameters:
// nFlags - Not used.
// point - Client coordinates of the mouse cursor.
//===================================================================
void GWCColorComponentEditCtrl::OnMouseMove(UINT nFlags, CPoint point) 
{
	// Two actions may be in progress when moving the mouse
	// 1. The mouse is dragged over the bottom color bar
	// 2. The mouse is dragged over the popup gradient color bar

	// If we drag the mouse over the bottom color bar
	if (m_dragInProgress)
	{
		// Calculate the new value based on the mouse location
		CRect colorRect = GetColorRect();
		if (point.x < colorRect.left)
			point.x = colorRect.left;
		else if (point.x > colorRect.right)
			point.x = colorRect.right;

		m_value = Round((double)(point.x-colorRect.left)/colorRect.Width()*255.);

		// Update the edit control with this value
		CString str;
		str.Format(_T("%d"),m_value);
		m_edit.SetWindowText(str);

		// Redraw the control to update the color bar
		RedrawWindow(colorRect,NULL);
	}

	// Or we drag the mouse over the popup gradient color bar ?
	else if (m_buttonPushed)
	{
		// Calculate the new value based on the mouse location
		CRect rect;
		m_gradientChooser->GetWindowRect(&rect);
		ClientToScreen(&point);
		int value = point.x - rect.left - 3;

		if (value < 0)
			value = 0;
		else if (value > 255)
			value = 255;

		m_value = value;

		// Update the edit control with this value
		CString str;
		str.Format(_T("%d"),m_value);
		m_edit.SetWindowText(str);

		// Force a redraw of everything
		m_edit.RedrawWindow();
		m_spin.RedrawWindow();
		RedrawWindow(&GetColorRect(),NULL,RDW_INVALIDATE|RDW_UPDATENOW);

		// Redraw the gradient bar because of the new cursor position
		m_gradientChooser->Invalidate();
	}

	CWnd::OnMouseMove(nFlags, point);
}

//=============================================================================
// Description:
// Called by the framework when the left mouse button is depressed, either over
// this control, or if the mouse was captured.
//
// Parameters:
// nFlags - Not used.
// point - Client coordinates of the mouse cursor.
//=============================================================================
void GWCColorComponentEditCtrl::OnLButtonUp(UINT nFlags, CPoint point) 
{
	// If there was a drag over a popup gradient bar
	if (m_buttonPushed)
	{
		// Hide the popup window
		m_gradientChooser->ShowWindow(SW_HIDE);

		m_buttonPushed = false;
		ReleaseCapture();

		// Redraw the button (no pushed effect)
		CRect buttonRect = GetButtonRect();
		RedrawWindow(&buttonRect,NULL);

		// Give the focus to the edit control and select the text
		m_edit.SetSel(0,-1);
		m_edit.SetFocus();
	}

	// or there was a drag over the bottom color bar ?
	else if (m_dragInProgress)
	{
		m_dragInProgress = false;
		ReleaseCapture();
	}

	CWnd::OnLButtonUp(nFlags, point);
}

//=================================================================
// Description:
// Returns the rectangle of the right button in client coordinates.
//
// Returns:
// The rectangle of the right button inclient coordinates.
//=================================================================
CRect GWCColorComponentEditCtrl::GetButtonRect()
{
	CRect rect;
	GetClientRect(&rect);
	rect.bottom -= 4;
	rect.left = rect.right - 11;

	return rect;
}

//=====================================================================
// Description:
// Returns the rectangle of the bottom color bar in client coordinates.
//
// Returns:
// The rectangle of the bottom color bar  inclient coordinates.
//=====================================================================
CRect GWCColorComponentEditCtrl::GetColorRect()
{
	CRect rect;
	GetClientRect(&rect);
	rect.top = rect.bottom - 3;

	return rect;
}

//=============================================
// Description:
// Sets the color that the color bar displays.
//
// Parameters:
// hue - Hue component of the new color.
// sat - Saturation component of the new color.
// light - Light component of the new color.
//=============================================
void GWCColorComponentEditCtrl::SetColorBar(BYTE hue, BYTE sat, BYTE light)
{
	m_colorBar = GWCColorFunc::HSLtoRGB((double)hue/255.,(double)sat/255.,(double)light/255.);
	CRect colorRect = GetColorRect();
	RedrawWindow(colorRect,NULL);
}

//============================================
// Description:
// Sets the color that the color bar displays.
//
// Parameters:
// color - RGB components of the new color.
//============================================
void GWCColorComponentEditCtrl::SetColorBar(COLORREF color)
{
	m_colorBar = color;
	CRect colorRect = GetColorRect();
	RedrawWindow(colorRect,NULL);
}

//==========================================================
// Description:
// Returns the component value displayed by this control.
//
// Returns:
// The component value displayed by this control (0 .. 255).
//==========================================================
unsigned char GWCColorComponentEditCtrl::GetValue()
{
	return m_value;
}

//====================================================
// Description:
// Sets the component value displayed by this control.
//
// Parameters:
// value - The new cvalue to display in this control.
//====================================================
void GWCColorComponentEditCtrl::SetValue(unsigned char value)
{
	// By setting this variable, EN_UPDATE won't be processed in OnCommand
	m_externalEditUpdate = true;

	m_value = value;

	// Update the edit control
	CString str;
	str.Format(_T("%d"),m_value);
	UpdateEditValue(str);

	// Update the spin control
	m_spin.SetPos(m_value);

	if (m_hWnd)
		Invalidate();

	m_externalEditUpdate = false;
}

//=================================================
// Description:
// Called when the user clicks on the spin control.
//
// Parameters:
// nSBCode - A code equal to SB_THUMBPOSITION.
// nPos - Position of the spin, from 0 to 255.
// pScrollBar - Not used.
//=================================================
void GWCColorComponentEditCtrl::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	if (nSBCode == SB_THUMBPOSITION)
	{
		m_value = nPos;

		CString str;
		str.Format(_T("%d"),m_value);
		m_edit.SetWindowText(str);
		m_edit.RedrawWindow();

		m_spin.RedrawWindow();

		Invalidate();
	}

	CWnd::OnVScroll(nSBCode, nPos, pScrollBar);
}

//==============================================================================================
// Description:
// The framework calls this method when, among other thing, a child control send a notification.
// Here we are interested by EN_UPDATE sent by the edit control.
//
// Parameters:
// wParam - Identifier of the notification code.
// lParam - Handle of the window that sent the notification.
//
// Returns:
// Not used.
//==============================================================================================
BOOL GWCColorComponentEditCtrl::OnCommand(WPARAM wParam, LPARAM lParam) 
{
	if (!m_externalEditUpdate && ((HWND)lParam == m_edit.GetSafeHwnd()) && (HIWORD(wParam) == EN_UPDATE))
	{
		CString str;
		m_edit.GetWindowText(str);

		// Check character typed
		bool charOK = true;
		for(int i=0;i<str.GetLength();i++)
		{
			// If the character is invalid, we just replace the whole value with the old one
			if ((str[i] < '0') || (str[i] > '9'))
			{
				str.Format(_T("%d"),m_value);
				UpdateEditValue(str);
				charOK = false;
			}
		}

		// If the character typed is valid
		if (charOK)
		{
			// Find the corresponding int value
			int value;
			if (str.GetLength() == 0)
				value = 0;
			else
				_stscanf(str.GetBuffer(0),_T("%d"),&value);

			if ((value < 0) || (value > 255))	// Invalid value
			{
				// Restore the old value
				str.Format(_T("%d"),m_value);
				UpdateEditValue(str);
			}
			else
			{
				// Set the new value
				m_value = value;
				m_spin.SetPos(m_value);
				Invalidate();

				if (m_type == HUE)
					SetColorBar(m_value,255,127);

				if (m_ownerSet)
					m_ownerSet->OnComponentChanged(this);

				GetParent()->SendMessage(WM_GWCCOLORCOMPEDIT_CHANGED,m_value,GetDlgCtrlID());

				// Remove leading 0s
				if (!str.GetLength() || (str.GetLength() && (str[0] == '0')))
				{
					str.Format(_T("%d"),m_value);
					UpdateEditValue(str);
				}
			}
		}
	}

	return CWnd::OnCommand(wParam, lParam);
}

//============================================================================================
// Description:
// Called by a GWCColorComponentSet owner object when the value of a control of the same class
// has been changed, if they are related to the same underlying color.
//
// Parameters:
// hue - Hue component of the new color.
// sat - Saturation component of the new color.
// light - Light component of the new color.
//============================================================================================
void GWCColorComponentEditCtrl::OnColorChanged(BYTE hue, BYTE sat, BYTE light)
{
	OnColorChanged(GWCColorFunc::HSLtoRGB((double)hue/255.,(double)sat/255.,(double)light/255.));
}

//============================================================================================
// Description:
// Called by a GWCColorComponentSet owner object when the value of a control of the same class
// has been changed, if they are related to the same underlying color.
//
// Parameters:
// color - RGB components of the new color.
//============================================================================================
void GWCColorComponentEditCtrl::OnColorChanged(COLORREF color)
{
	// Update our internal value and change the color bar if necessary

	if (m_type == RED)
		SetValue(GetRValue(color));
	else if (m_type == GREEN)
		SetValue(GetGValue(color));
	else if (m_type == BLUE)
		SetValue(GetBValue(color));
	else if (m_type == HUE)
	{
		double h,s,l;
		GWCColorFunc::RGBtoHSL(color,&h,&s,&l);
		SetValue(Round(h * 255.));
		SetColorBar(Round(h * 255.),255,127);
	}
	else if (m_type == SAT)
	{
		double h,s,l;
		GWCColorFunc::RGBtoHSL(color,&h,&s,&l);
		SetValue(Round(s * 255.));
		SetColorBar(Round(h * 255.),255,127);
	}
	else if (m_type == LIGHT)
	{
		double h,s,l;
		GWCColorFunc::RGBtoHSL(color,&h,&s,&l);
		SetValue(Round(l * 255.));
		SetColorBar(Round(h * 255.),255,127);
	}

	RedrawWindow();
}

//========================================================================================
// Description:
// Returns the component type of this control. Can be RED, GREEN, BLUE, HUE, SAT or LIGHT.
//
// Returns:
// The component type of this control.
//========================================================================================
int GWCColorComponentEditCtrl::GetType()
{
	return m_type;
}

//=====================================================================================
// Description:
// Sets the component type of this control. Can be RED, GREEN, BLUE, HUE, SAT or LIGHT.
//
// Parameters:
// type - The component type of this control.
//=====================================================================================
void GWCColorComponentEditCtrl::SetType(int type)
{
	m_type = type;

	// Update the color bar if necessary
	if (m_type == RED)
		SetColorBar(RGB(255,0,0));
	else if (m_type == GREEN)
		SetColorBar(RGB(0,255,0));
	else if (m_type == BLUE)
		SetColorBar(RGB(0,0,255));
}

//=======================================================================
// Description:
// Sets the value displayed in the edit control.
//
// Parameters:
// str - A string containing the new integer value to display (0 .. 255).
//=======================================================================
void GWCColorComponentEditCtrl::UpdateEditValue(CString& str)
{
	// By setting this variable, EN_UPDATE won't be processed in OnCommand
	m_externalEditUpdate = true;

	m_edit.SetWindowText(str);

	m_externalEditUpdate = false;
}

//===================================================================================
// Description:
// Sets the owner GWCColorComponentSet that groups several color components together.
//
// Parameters:
// owner - A pointer to the owner GWCColorComponentSet.
//===================================================================================
void GWCColorComponentEditCtrl::SetOwnerSet(GWCColorComponentSet* owner)
{
	m_ownerSet = owner;
}

//======================================================================================
// Description:
// Returns the owner GWCColorComponentSet that groups several color components together.
//
// Returns:
// A pointer to the owner GWCColorComponentSet.
//======================================================================================
GWCColorComponentSet* GWCColorComponentEditCtrl::GetOwnerSet()
{
	return m_ownerSet;
}

//============================================================
// Description:
// Called by the framework when this control looses the focus.
//
// Parameters:
// pNewWnd - A pointer to the window that gained the focus.
//============================================================
void GWCColorComponentEditCtrl::OnKillFocus(CWnd* pNewWnd) 
{
	CWnd::OnKillFocus(pNewWnd);

	// If the gradient color bar is visible then hide it
	if (m_gradientChooser)
	{
		m_gradientChooser->ShowWindow(SW_HIDE);

		m_buttonPushed = false;
		ReleaseCapture();

		CRect buttonRect = GetButtonRect();
		RedrawWindow(&buttonRect,NULL);
	}
}

//===========================================================
// Description:
// Called by the framework when this control gains the focus.
//
// Parameters:
// pOldWnd - A pointer to the window that lose the focus.
//===========================================================
void GWCColorComponentEditCtrl::OnSetFocus(CWnd* pOldWnd) 
{
	CWnd::OnSetFocus(pOldWnd);

	// If the right button has not been clicked to gain the focus
	if (m_buttonFocus == false)
	{
		// Give the focus to the edit control and select the text
		m_edit.SetFocus();
		m_edit.SetSel(0,-1);
	}
	else
		m_buttonFocus = false;
}
